# Run jeni tests, incorporating Python 2 / Python 3 specifics.

import sys
import unittest

import test_jeni
try:
    unittest.main(module=test_jeni)
except SystemExit:
    pass

if sys.version_info < (3,):
    import test_jeni_python2
    unittest.main(module=test_jeni_python2)
else:
    import test_jeni_python3
    unittest.main(module=test_jeni_python3)



from collections import OrderedDict as odict
from decimal import Decimal
from fractions import Fraction
import sys
import unittest

import jeni


class BasicInjector(jeni.Injector):
    pass


@BasicInjector.provider('hello')
class HelloProvider(jeni.Provider):
    def get(self, name=None):
        if name is None:
            name = 'world'
        return 'Hello, {}!'.format(name)


@BasicInjector.factory('eggs')
def eggs():
    return 'eggs!'


@BasicInjector.factory('echo')
def echo(name=None):
    return name


@BasicInjector.factory('space')
def space():
    return {}


BasicInjector.value('zero', 0)


@BasicInjector.provider('answer')
def answer():
    yield 42


@BasicInjector.provider('spam', name=True)
def spam():
    count_str = yield 'spam'
    while True:
        count_str = yield 'spam' * int(count_str)


class WithOrderedSpace(jeni.Injector):
    pass


@WithOrderedSpace.factory('space')
def ospace():
    return odict()


class SubInjector(BasicInjector):
    pass


@SubInjector.provider('answer')
def sub_answer():
    yield 'No one knows.'


class BoringInjector(jeni.Injector):
    pass


@BoringInjector.factory('hello')
@BoringInjector.factory('eggs')
@BoringInjector.factory('answer')
@BoringInjector.factory('spam')
def be_boring(name=None):
    if name is None:
        name = 'this injector'
    return '{} is boring'.format(name)


class BasicInjectorTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_provider_class(self):
        self.assertEqual('Hello, world!', self.injector.get('hello'))
        self.assertEqual('Hello, world!', self.injector.get('hello'))
        self.assertEqual('Hello, thing!', self.injector.get('hello:thing'))
        self.assertEqual('Hello, thing!', self.injector.get('hello:thing'))

    def test_factory(self):
        self.assertEqual('eggs!', self.injector.get('eggs'))
        self.assertEqual('eggs!', self.injector.get('eggs'))

        # This factory does not accept name.
        self.assertRaises(TypeError, self.injector.get, 'eggs:thing')

    def test_factory_identity(self):
        space1 = self.injector.get('space')
        space2 = self.injector.get('space')
        self.assertIs(space1, space2)

    def test_factory_with_name(self):
        self.assertEqual(None, self.injector.get('echo'))
        self.assertEqual('foo', self.injector.get('echo:foo'))

    def test_value(self):
        self.assertEqual(0, self.injector.get('zero'))
        self.assertEqual(0, self.injector.get('zero'))

        # This does not accept name.
        self.assertRaises(TypeError, self.injector.get, 'zero:thing')

    def test_generator(self):
        self.assertEqual(42, self.injector.get('answer'))
        self.assertEqual(42, self.injector.get('answer'))

        # This generator does not accept name.
        self.assertRaises(TypeError, self.injector.get, 'answer:thing')

    def test_generator_with_name(self):
        self.assertEqual('spam', self.injector.get('spam'))
        self.assertEqual('spam', self.injector.get('spam'))
        self.assertEqual('spamspam', self.injector.get('spam:2'))
        self.assertEqual('spamspamspam', self.injector.get('spam:3'))
        self.assertEqual('spamspamspamspam', self.injector.get('spam:4'))
        self.assertEqual('spam', self.injector.get('spam'))

        self.assertRaises(ValueError, self.injector.get, 'spam:forty')
        self.assertEqual('spam', self.injector.get('spam'))

    def test_not_registered(self):
        self.assertRaises(LookupError, self.injector.get, 'nothing')

    def test_provider_class_registration(self):
        class SubInjector(BasicInjector):
            pass
        SubInjector.provider('hello2', HelloProvider)
        injector = SubInjector()
        self.assertEqual('Hello, world!', injector.get('hello2'))
        self.assertEqual('Hello, world!', injector.get('hello'))

    def test_factory_registration(self):
        class SubInjector(BasicInjector):
            pass
        SubInjector.factory('eggs2', eggs)
        injector = SubInjector()
        self.assertEqual('eggs!', injector.get('eggs2'))
        self.assertEqual('eggs!', injector.get('eggs'))


class InjectSelfTestCase(unittest.TestCase):
    def test_provide_self_default(self):
        self.injector = jeni.Injector()
        self.assertRaises(LookupError, self.injector.get, 'injector')

    def test_provide_self_true(self):
        self.injector = jeni.Injector(provide_self=True)
        self.assertEqual(self.injector, self.injector.get('injector'))

    def test_provide_self_false(self):
        self.injector = jeni.Injector(provide_self=False)
        self.assertRaises(LookupError, self.injector.get, 'injector')


class SubInjectorTestCase(BasicInjectorTestCase):
    def setUp(self):
        self.injector = SubInjector()

    def test_generator(self):
        self.assertEqual('No one knows.', self.injector.get('answer'))
        self.assertEqual('No one knows.', self.injector.get('answer'))

        # This generator does not accept name.
        self.assertRaises(TypeError, self.injector.get, 'answer:thing')


class InjectorSubTestCase(BasicInjectorTestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_subinjector(self):
        subinjector = self.injector.sub()
        self.assertIsInstance(subinjector, jeni.Injector)

    def test_subinjector_inheritance(self):
        injector_class = type(self.injector)
        subinjector_class = type(self.injector.sub())
        self.assertTrue(issubclass(subinjector_class, injector_class))

    def test_subinjector_isolation(self):
        space = self.injector.get('space')
        subinjector = self.injector.sub()
        sub_space = subinjector.get('space')
        self.assertIsNot(space, sub_space)

    def test_subinjector_local_values(self):
        subinjector = self.injector.sub(zero=0.0)
        self.assertIsInstance(subinjector.get('zero'), float)

        local_values = dict(zero=0.0)
        subinjector = self.injector.sub(local_values)
        self.assertIsInstance(subinjector.get('zero'), float)

        more_local_values = dict(zero=Fraction(0, 1))
        subinjector = self.injector.sub(local_values, more_local_values)
        self.assertIsInstance(subinjector.get('zero'), Fraction)

        subinjector = self.injector.sub(
                local_values, more_local_values, zero=Decimal('0'))
        self.assertIsInstance(subinjector.get('zero'), Decimal)

    def test_subinjector_mixins(self):
        subinjector = self.injector.sub(WithOrderedSpace)
        injector_class = type(self.injector)
        subinjector_class = type(self.injector.sub())
        self.assertTrue(issubclass(subinjector_class, injector_class))
        self.assertIsInstance(subinjector, WithOrderedSpace)
        self.assertIsInstance(subinjector.get('space'), odict)


class BasicInjectorAnnotationTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()
        @jeni.annotate('hello', 'hello:x', 'eggs', 'answer', 'spam', 'spam:2')
        def fn(hello, hello_x, eggs, answer, spam, spam2):
            self.assertEqual('Hello, world!', hello)
            self.assertEqual('Hello, x!', hello_x)
            self.assertEqual('eggs!', eggs)
            self.assertEqual(42, answer)
            self.assertEqual('spam', spam)
            self.assertEqual('spamspam', spam2)
            return 0
        self.fn = fn

    def test_apply(self):
        self.assertEqual(0, self.injector.apply(self.fn))

    def test_partial(self):
        fn = self.injector.partial(self.fn)
        self.assertEqual(0, fn())
        self.assertEqual(0, fn())

    def test_eager_partial(self):
        fn = self.injector.eager_partial(self.fn)
        self.assertEqual(0, fn())
        self.assertEqual(0, fn())

    def test_not_registered(self):
        @jeni.annotate('hello', 'nothing')
        def fn(hello, nothing):
            raise RuntimeError('UnsetError should raise before this on apply.')
        self.assertRaises(RuntimeError, fn, 'hello', None)
        self.assertRaises(LookupError, self.injector.apply, fn)

    def test_no_annotation(self):
        def fn():
            "unused"
        self.assertRaises(AttributeError, self.injector.apply, fn)

    def test_annotate_without_annotations(self):
        # This instrumentation works in both Python 2 & Python 3.
        def fn(hello):
            "unused"
        fn.__annotations__ = {'hello': 'hello:world'}
        jeni.annotate(fn)
        self.assertTrue(jeni.annotate.has_annotations(fn))

    def test_annotate_fn_name_keyword(self):
        # Test that annotation of 'fn' is not accidentally reserved by jeni.
        def foo(fn=None):
            "unused"
        decorator = jeni.annotate(fn='function')

        try:
            decorator(foo)
            raise TypeError('code coverage support')
        except TypeError:
            _, err, _ = sys.exc_info()
            msg = "Annotation cannot support argument named 'fn'."
            if 'coverage' not in str(err): self.fail(msg)


class SubInjectorAnnotationTestCase(BasicInjectorAnnotationTestCase):
    def setUp(self):
        self.injector = SubInjector()
        @jeni.annotate('hello', 'hello:x', 'eggs', 'answer', 'spam', 'spam:2')
        def fn(hello, hello_x, eggs, answer, spam, spam2):
            self.assertEqual('Hello, world!', hello)
            self.assertEqual('Hello, x!', hello_x)
            self.assertEqual('eggs!', eggs)
            self.assertEqual('No one knows.', answer)
            self.assertEqual('spam', spam)
            self.assertEqual('spamspam', spam2)
            return 0
        self.fn = fn


class BoringAnnotationTestCase(BasicInjectorAnnotationTestCase):
    def setUp(self):
        self.injector = BoringInjector()
        @jeni.annotate('hello', 'hello:x', 'eggs', 'answer', 'spam', 'spam:2')
        def fn(hello, hello_x, eggs, answer, spam, spam2):
            self.assertEqual('this injector is boring', hello)
            self.assertEqual('x is boring', hello_x)
            self.assertEqual('this injector is boring', eggs)
            self.assertEqual('this injector is boring', answer)
            self.assertEqual('this injector is boring', spam)
            self.assertEqual('2 is boring', spam2)
            return 0
        self.fn = fn


class ApplyScenariosTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

        @jeni.annotate('hello')
        def fn_annotated(hello, *a, **kw):
            return hello, a, kw
        self.fn_annotated = fn_annotated

        def fn_not_annotated(*a, **kw):
            return a, kw
        self.fn_not_annotated = fn_not_annotated

        @jeni.annotate(a='doesnotexist1', b='doesnotexist2', c='hello')
        def fn_with_lookup_error(a, b, c=None):
            return a, b, c
        self.fn_with_lookup_error = fn_with_lookup_error

    def test_apply(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.apply(self.fn_annotated))
        self.assertRaises(
            AttributeError,
            self.injector.apply, self.fn_not_annotated)

    def test_additional_arguments_apply(self):
        self.assertEqual(
            ('Hello, world!', ('a', 'b'), {'letter': 'c'}),
            self.injector.apply(self.fn_annotated, 'a', 'b', letter='c'))
        self.assertRaises(
            AttributeError,
            self.injector.apply, self.fn_not_annotated, 'a', 'b', letter='c')

    def test_apply_regardless(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.apply_regardless(self.fn_annotated))
        self.assertEqual(
            ((), {}),
            self.injector.apply_regardless(self.fn_not_annotated))

    def test_additional_arguments_apply_regardless(self):
        self.assertEqual(
            ('Hello, world!', ('a', 'b'), {'letter': 'c'}),
            self.injector.apply_regardless(
                self.fn_annotated, 'a', 'b', letter='c'))
        self.assertEqual(
            (('a', 'b'), {'letter': 'c'}),
            self.injector.apply_regardless(
                self.fn_not_annotated, 'a', 'b', letter='c'))

    def test_partial(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.partial(self.fn_annotated)())
        self.assertRaises(
            AttributeError,
            self.injector.partial, self.fn_not_annotated)

    def test_eager_partial(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.eager_partial(self.fn_annotated)())
        self.assertRaises(
            AttributeError,
            self.injector.eager_partial, self.fn_not_annotated)

    def test_additional_arguments_partial(self):
        self.assertEqual(
            ('Hello, world!', ('a', 'b'), {'letter': 'c'}),
            self.injector.partial(self.fn_annotated, 'a', 'b', letter='c')())
        self.assertRaises(
            AttributeError,
            self.injector.partial, self.fn_not_annotated, 'a', 'b', letter='c')

    def test_partial_as_maybe(self):
        self.assertEqual(
            ('a', 'b', 'Hello, world!'),
            self.injector.partial(self.fn_with_lookup_error, 'a', 'b')())
        self.assertEqual(
            ('a', 'b', 'Hello, world!'),
            self.injector.partial(self.fn_with_lookup_error)('a', 'b'))

    def test_eager_partial_as_maybe(self):
        self.assertEqual(
            ('a', 'b', 'Hello, world!'),
            self.injector.eager_partial(self.fn_with_lookup_error, 'a', 'b')())
        self.assertEqual(
            ('a', 'b', 'Hello, world!'),
            self.injector.eager_partial(self.fn_with_lookup_error)('a', 'b'))

    def test_partial_regardless(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.partial_regardless(self.fn_annotated)())
        self.assertEqual(
            ((), {}),
            self.injector.partial_regardless(self.fn_not_annotated)())

    def test_eager_partial_regardless(self):
        self.assertEqual(
            ('Hello, world!', (), {}),
            self.injector.eager_partial_regardless(self.fn_annotated)())
        self.assertEqual(
            ((), {}),
            self.injector.eager_partial_regardless(self.fn_not_annotated)())

    def test_additional_arguments_partial_regardless(self):
        self.assertEqual(
            ('Hello, world!', ('a', 'b'), {'letter': 'c'}),
            self.injector.partial_regardless(
                self.fn_annotated, 'a', 'b', letter='c')())
        self.assertEqual(
            (('a', 'b'), {'letter': 'c'}),
            self.injector.partial_regardless(
                self.fn_not_annotated, 'a', 'b', letter='c')())


@jeni.annotate('spam', 'eggs')
def spam_eggs(spam, eggs, name=None):
    if name is not None:
        return ' '.join((spam, eggs, name))
    return ' '.join((spam, eggs))


class AnnotatedInitProvider(jeni.Provider):
    @jeni.annotate('spam:4')
    def __init__(self, spam):
        self.foo = None
        self.spam = spam

    def get(self):
        return self

    def close(self):
        self.foo = self.spam


class AnnotatedProviderTestCase(unittest.TestCase):
    def test_spam_eggs(self):
        class Injector(BasicInjector):
            pass
        Injector.factory('dish', spam_eggs)
        injector = Injector()
        self.assertEqual(('spam eggs!'), injector.get('dish'))
        self.assertEqual(('spam eggs! eel'), injector.get('dish:eel'))

    def test_annotated_generator(self):
        class Injector(BasicInjector):
            pass
        @Injector.provider('spammish')
        @jeni.annotate('spam')
        def generator(spam):
            yield spam
        injector = Injector()
        self.assertEqual('spam', injector.get('spammish'))

    def test_annotated_init(self):
        class Injector(BasicInjector):
            pass
        Injector.provider('annotated_init', AnnotatedInitProvider)
        injector = Injector()
        provider = injector.get('annotated_init')
        self.assertEqual(None, provider.foo)
        injector.close()
        self.assertEqual('spamspamspamspam', provider.foo)


class TestCycles(unittest.TestCase):
    def setUp(self):
        class Injector(jeni.Injector): pass

        @Injector.factory('one')
        @jeni.annotate('three')
        def three_minus_2(three):
            return not three-2

        @Injector.factory('two')
        @jeni.annotate('one')
        def one_times_2(one):
            return not one*2

        @Injector.factory('three')
        @jeni.annotate('two')
        def two_plus_1(two):
            return not two+1

        self.injector = Injector()

        # Call functions to simplify coverage.
        three_minus_2(3)
        one_times_2(1)
        two_plus_1(2)

    def test_cycles(self):
        with self.assertRaises(jeni.DependencyCycleError) as raises:
            self.injector.get('one')

        self.assertEqual(raises.exception.notes, (
                ('one', None),
                ('three', None),
                ('two', None),
                ('one', None)))


class WrapsTestCase(unittest.TestCase):
    def setUp(self):
        @jeni.annotate('spam')
        def foo(spam):
            "Foo the spam."
            return spam
        self.foo = foo
        @jeni.wraps(foo)
        def wrapper(spam):
            return ('spam', spam)
        self.wrapper = wrapper
        self.injector = BasicInjector()

    def test_has_annotations(self):
        self.assertTrue(jeni.annotate.has_annotations(self.wrapper))

    def test_name_and_doc(self):
        self.assertEqual('foo', self.wrapper.__name__)
        self.assertEqual('Foo the spam.', self.wrapper.__doc__)

    def test_injection(self):
        self.assertEqual('spam', self.injector.apply(self.foo))
        self.assertEqual(('spam', 'spam'), self.injector.apply(self.wrapper))


class MaybeTestCase(unittest.TestCase):
    def setUp(self):
        self.note = jeni.maybe('the_real_note')

    def test_maybe(self):
        self.assertEqual((jeni.MAYBE, 'the_real_note'), self.note)


class PartialNoteTestCase(unittest.TestCase):
    def setUp(self):
        self.fn = lambda a, b, c: None
        self.note = jeni.partial(self.fn)
        self.note_with_args = jeni.partial(self.fn, 'a', 'bee', c='cee')

    def test_partial(self):
        # (PARTIAL, (function, args, kwargs_items_tuple))
        self.assertEqual((jeni.PARTIAL, (self.fn, (), ())), self.note)

    def test_partial_with_args(self):
        self.assertEqual(
            (jeni.PARTIAL, (self.fn, ('a', 'bee'), (('c', 'cee'),))),
            self.note_with_args)


class EagerPartialNoteTestCase(unittest.TestCase):
    def setUp(self):
        self.fn = lambda a, b, c: None
        self.note = jeni.eager_partial(self.fn)
        self.note_with_args = jeni.eager_partial(self.fn, 'a', 'bee', c='cee')

    def test_partial(self):
        # (EAGER_PARTIAL, (function, args, kwargs_items_tuple))
        self.assertEqual((jeni.EAGER_PARTIAL, (self.fn, (), ())), self.note)

    def test_partial_with_args(self):
        self.assertEqual(
            (jeni.EAGER_PARTIAL, (self.fn, ('a', 'bee'), (('c', 'cee'),))),
            self.note_with_args)


@BasicInjector.factory('error')
def error():
    raise jeni.UnsetError()


@jeni.annotate('error')
def unset_arg(unused):
    raise RuntimeError('UnsetError should raise before this on apply.')


@jeni.annotate('hello', unused='error')
def unset_kwarg(hello, unused=None):
    raise RuntimeError('UnsetError should raise before this on apply.')


@jeni.annotate('hello', unused=jeni.annotate.maybe('error'))
def unset_maybe_kwarg(hello, unused=None):
    return unused


@jeni.annotate('hello', unused=jeni.annotate.maybe('notprovided'))
def notprovided_maybe_kwarg(hello, unused=None):
    return unused


class UnsetArgumentTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_unset_arg(self):
        self.assertRaises(RuntimeError, unset_arg, None)
        self.assertRaises(jeni.UnsetError, self.injector.apply, unset_arg)

    def test_unset_kwarg(self):
        self.assertRaises(RuntimeError, unset_kwarg, None)
        self.assertRaises(jeni.UnsetError, self.injector.apply, unset_kwarg)

    def test_unset_maybe_kwarg(self):
        self.assertIs(None, self.injector.apply(unset_maybe_kwarg))

    def test_notprovided_maybe_kwarg(self):
        self.assertIs(None, self.injector.apply(notprovided_maybe_kwarg))


class UnsetInjector(jeni.Injector):
    pass


@UnsetInjector.factory('unset_msg')
def unset_msg_factory():
    raise jeni.UnsetError('exception message')


@UnsetInjector.factory('unset_no_msg')
def unset_no_msg_factory():
    raise jeni.UnsetError()


class UnsetErrorMessageTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = UnsetInjector()

    def get_error(self, note):
        err = None
        try:
            self.injector.get(note)
        except jeni.UnsetError:
            _, err, _ = sys.exc_info()
        self.assertIsNotNone(err)
        return err

    def test_unset_msg(self):
        err = self.get_error('unset_msg')
        self.assertEqual(str(err), "exception message: 'unset_msg'")
        self.assertEqual(err.note, 'unset_msg')

    def test_unset_no_msg(self):
        err = self.get_error('unset_no_msg')
        self.assertEqual(str(err), "'unset_no_msg'")
        self.assertEqual(err.note, 'unset_no_msg')


def hello_simple():
    return 'wat'


@jeni.annotate('hello:partial')
def hello_partial(hello):
    return hello


@jeni.annotate('hello:again', jeni.annotate.partial(hello_partial))
def hello_again_partial(hello, fn, **keywords):
    return hello, fn(), keywords


@jeni.annotate('hello:again', jeni.annotate.partial_regardless(hello_partial))
def hello_again_partial_regardless_annotated(hello, fn, **keywords):
    return hello, fn(), keywords


@jeni.annotate('hello:again', jeni.annotate.partial_regardless(hello_simple))
def hello_again_partial_regardless_unannotated(hello, fn, **keywords):
    return hello, fn(), keywords


@jeni.annotate('hello:again', jeni.annotate.eager_partial(hello_partial))
def hello_again_eager_partial(hello, fn, **keywords):
    return hello, fn(), keywords


@jeni.annotate('hello:again', jeni.annotate.eager_partial_regardless(hello_partial))
def hello_again_eager_partial_regardless_annotated(hello, fn, **keywords):
    return hello, fn(), keywords


@jeni.annotate('hello:again', jeni.annotate.eager_partial_regardless(hello_simple))
def hello_again_eager_partial_regardless_unannotated(hello, fn, **keywords):
    return hello, fn(), keywords


class InjectPartialTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_partial_injection(self):
        self.assertEqual(
            ('Hello, again!', 'Hello, partial!', {'c': 'cee'}),
            self.injector.apply(hello_again_partial, c='cee'))

    def test_partial_regardless_injection(self):
        self.assertEqual(
            ('Hello, again!', 'Hello, partial!', {'c': 'cee'}),
            self.injector.apply(
                    hello_again_partial_regardless_annotated, c='cee'))
        self.assertEqual(
            ('Hello, again!', 'wat', {'c': 'cee'}),
            self.injector.apply(
                    hello_again_partial_regardless_unannotated, c='cee'))

    def test_eager_partial_injection(self):
        self.assertEqual(
            ('Hello, again!', 'Hello, partial!', {'c': 'cee'}),
            self.injector.apply(hello_again_eager_partial, c='cee'))

    def test_eager_partial_regardless_injection(self):
        self.assertEqual(
            ('Hello, again!', 'Hello, partial!', {'c': 'cee'}),
            self.injector.apply(
                    hello_again_eager_partial_regardless_annotated, c='cee'))
        self.assertEqual(
            ('Hello, again!', 'wat', {'c': 'cee'}),
            self.injector.apply(
                    hello_again_eager_partial_regardless_unannotated, c='cee'))


class LazyPartialTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_partial(self):
        note = 'hello:partial'
        self.assertEqual(0, self.injector.stats[note])
        fn = self.injector.partial(hello_partial)
        self.assertEqual(0, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])

    def test_eager_partial(self):
        note = 'hello:partial'
        self.assertEqual(0, self.injector.stats[note])
        fn = self.injector.eager_partial(hello_partial)
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])

    def test_partial_injection(self):
        note = 'hello:partial'
        self.assertEqual(0, self.injector.stats[note])
        fn = self.injector.partial(hello_again_partial)
        self.assertEqual(0, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])

    def test_eager_partial_injection(self):
        note = 'hello:partial'
        self.assertEqual(0, self.injector.stats[note])
        fn = self.injector.eager_partial(hello_again_eager_partial)
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])
        fn()
        self.assertEqual(1, self.injector.stats[note])


class CloseMe(object):
    # List of all closed instances for use in test inspection.
    closed_items = []

    def __init__(self, note):
        self.closed = None
        self.note = note

    def open(self):
        self.closed = False

    def close(self):
        assert not self.closed, '{!r} already closed'.format(self)
        self.closed_items.append(self)
        self.closed = True


class CloseTestInjector(jeni.Injector):
    pass


@CloseTestInjector.provider('via_class')
class ClosingProvider(jeni.Provider):
    def __init__(self):
        self.thing = CloseMe('via_class')

    def get(self, name=None):
        if self.thing.closed is None:
            self.thing.open()
        return self.thing

    def close(self):
        self.thing.close()


@CloseTestInjector.provider('via_generator')
def closing_generator():
    thing = CloseMe('via_generator')
    thing.open()
    yield thing
    thing.close()


@CloseTestInjector.provider('via_generator_with_name', name=True)
def closing_generator_with_name():
    # Name is not actually used, but generator accepts name.
    try:
        thing = CloseMe('via_generator_with_name')
        thing.open()
        yield thing
        while True:
            yield thing
    finally:
        thing.close()


@CloseTestInjector.factory('unset_thing')
def unset_thing_factory():
    return CloseMe('unset')


@CloseTestInjector.provider('unset')
class UnsetProvider(jeni.Provider):
    @jeni.annotate('unset_thing')
    def __init__(self, thing):
        self.closed = False
        self.thing = thing

    def get(self):
        raise jeni.UnsetError()

    def close(self):
        assert not self.closed, '{!r} already closed'.format(self)
        self.closed = True
        self.thing.close()


@CloseTestInjector.provider('annotated_close')
class AnnotatedCloseProvider(jeni.Provider):
    def get(self):
        return self

    @jeni.annotate('echo:bar')
    def close(self, echo):
        "This should raise TypeError."


CloseTestInjector.factory('echo', echo)


class ClosingTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = CloseTestInjector()

    def test_multiple_close(self):
        self.injector.close()
        self.assertRaises(RuntimeError, self.injector.close)
        self.assertRaises(RuntimeError, self.injector.close)

    def assert_with_note(self, note):
        thing = self.injector.get(note)
        self.assertIs(thing, self.injector.get(note))
        self.assertEqual(False, thing.closed)
        self.injector.close()
        self.assertEqual(True, thing.closed)

    def test_close_via_class(self):
        self.assert_with_note('via_class')

    def test_close_via_generator(self):
        self.assert_with_note('via_generator')

    def test_close_via_generator_with_name(self):
        self.assert_with_note('via_generator_with_name')

    def test_close_via_generator_with_name_name(self):
        self.assert_with_note('via_generator_with_name:name')

    def _test_close_order(self):
        injector = CloseTestInjector()
        num_prev_closed_items = len(CloseMe.closed_items)
        notes = ['via_generator', 'via_class', 'via_generator_with_name']
        close_order = list(reversed(notes))
        injector.get('echo') # Add a non-closing provider for good measure.
        for note in notes:
            injector.get(note)
        injector.get('via_class') # Get again; should not impact order.
        injector.close()
        self.assertEqual(
            close_order,
            [x.note for x in CloseMe.closed_items[num_prev_closed_items:]])

    def test_close_order(self):
        # Test multiple times given failed test is random order.
        for x in range(100):
            self._test_close_order()

    def test_unset_closed(self):
        thing = self.injector.get('unset_thing')
        self.assertRaises(jeni.UnsetError, self.injector.get, 'unset')
        self.injector.close()
        self.assertTrue(thing.closed)

    def test_cannot_get_after_close(self):
        self.assertEqual('thing', self.injector.get('echo:thing'))
        self.injector.close()
        self.assertRaises(RuntimeError, self.injector.get, 'echo:thing')

    def test_annotated_close(self):
        self.injector.get('annotated_close')
        # Injector.close does not apply annotations.
        self.assertRaises(TypeError, self.injector.close)

    def test_multiple_get_then_close(self):
        self.injector.get('via_class:foo')
        self.injector.get('via_class:bar')
        self.injector.close() # AssertionError likely if this fails.


class InjectorStatsTestCase(unittest.TestCase):
    def setUp(self):
        self.injector = BasicInjector()

    def test_empty_stats(self):
        self.assertEqual({}, self.injector.stats)

    def test_a_few_calls(self):
        stats = {
            'hello': 1,
            'hello:thing': 1,
            'eggs': 2,
            (jeni.PARTIAL_REGARDLESS, (eggs, (), ())): 1,
        }
        self.injector.get('eggs')
        self.injector.get('hello')
        self.injector.get('hello:thing')
        self.injector.get('eggs')
        self.assertEqual(stats, self.injector.stats)

    def test_many_calls(self):
        stats = {
            'hello': 10,
            'hello:thing': 15,
            'eggs': 21,
            (jeni.PARTIAL_REGARDLESS, (eggs, (), ())): 1,
        }
        for _ in range(10):
            self.injector.get('hello')
        for _ in range(21):
            self.injector.get('eggs')
        for _ in range(15):
            self.injector.get('hello:thing')
        self.assertEqual(stats, self.injector.stats)


class ContextManagerTestCase(unittest.TestCase):
    def test_with_block(self):
        with CloseTestInjector() as injector:
            thing = injector.get('via_class')
            self.assertEqual(False, thing.closed)
        self.assertEqual(True, thing.closed)

    def test_enter_exit(self):
        injector = CloseTestInjector().enter()
        thing = injector.get('via_class')
        self.assertEqual(False, thing.closed)
        injector.exit()
        self.assertEqual(True, thing.closed)


class TestGeneratorProvider(unittest.TestCase):
    def test_generator(self):
        def fn():
            yield 42
        provider = jeni.GeneratorProvider(fn)
        self.assertEqual(42, provider.get())
        self.assertEqual(42, provider.get())
        self.assertRaises(TypeError, provider.get, name='name')

    def test_construction_error(self):
        def fn():
            "not a generator"
        self.assertRaises(TypeError, jeni.GeneratorProvider, fn)

    def test_init_no_error(self):
        def fn():
            yield 42
        provider = jeni.GeneratorProvider(fn)
        provider.get()
        provider.close()

    def test_unyielding_generator(self):
        def fn(work=False):
            if work:
                yield 'foo'
        self.assertEqual(['foo'], list(fn(work=True)))
        self.assertRaises(RuntimeError, jeni.GeneratorProvider, fn)

    def test_generator_with_broken_name_support(self):
        def fn():
            yield 42
        provider = jeni.GeneratorProvider(fn, support_name=True)
        self.assertEqual(42, provider.get())
        self.assertRaises(RuntimeError, provider.get, name='name')

    def test_generator_which_keeps_yielding(self):
        def fn():
            yield 'one'; yield 'two'
        provider = jeni.GeneratorProvider(fn)
        self.assertRaises(RuntimeError, provider.close)


class NonStringNoteTestCase(unittest.TestCase):
    class TestInjector(jeni.Injector):
        pass

    def setUp(self):
        self.Injector = self.TestInjector
        self.injector = self.Injector()

    def test_object(self):
        note = object()
        self.Injector.provider(note, HelloProvider)
        self.assertEqual('Hello, world!', self.injector.get(note))

    def test_tuple(self):
        note = ('hello', 'name')
        self.Injector.provider(note, HelloProvider)
        self.assertEqual('Hello, name!', self.injector.get(note))

    def test_tuple_too_small(self):
        note = ('hello',)
        self.assertRaises(
            ValueError, self.Injector.provider, note, HelloProvider)

    def test_tuple_too_large(self):
        note = ('hello', 'name', 'bogus')
        self.assertRaises(
            ValueError, self.Injector.provider, note, HelloProvider)


class MoreAnnotationTests(unittest.TestCase):
    def test_multiple_annotations(self):
        @jeni.annotate('foo', 'bar')
        def fn(foo, bar):
            "unused"
        decorator = jeni.annotate('baz', 'quux')
        self.assertRaises(AttributeError, decorator, fn)

    def test_method_annotation(self):
        class X(object):
            @jeni.annotate('spam', 'eggs')
            def eat(self, spam, eggs):
                return spam + eggs
        injector = BasicInjector()
        x = X()
        self.assertEqual('spameggs!', injector.apply(x.eat))

    def test_classmethod_annotation(self):
        class X(object):
            @jeni.annotate('spam', 'eggs')
            @classmethod
            def eat(cls, spam, eggs):
                return spam + eggs
        injector = BasicInjector()
        self.assertEqual('spameggs!', injector.apply(X.eat))


class TestBrokenProvider(unittest.TestCase):
    class TestInjector(jeni.Injector):
        pass

    def setUp(self):
        self.Injector = self.TestInjector
        self.injector = self.Injector()

    def subclass(self):
        class BadSubclass(jeni.Provider):
            pass
        return BadSubclass

    def test_subclass_meta(self):
        cls = self.subclass()
        self.assertRaises(TypeError, cls)


class TestInjectorProxy(unittest.TestCase):
    def setUp(self):
        self.x = jeni.InjectorProxy(BasicInjector())

    def test_getattr(self):
        self.assertEqual('Hello, world!', self.x.hello)
        self.assertEqual('Hello, thing!', getattr(self.x, 'hello:thing'))

    def test_unset_getattr(self):
        def _test():
            self.x.error
        self.assertRaises(jeni.UnsetError, _test)

    def test_getitem(self):
        self.assertEqual('Hello, world!', self.x['hello'])
        self.assertEqual('Hello, thing!', self.x['hello:thing'])

    def test_unset_getitem(self):
        def _test():
            self.x['error']
        self.assertRaises(jeni.UnsetError, _test)

    def test_in(self):
        self.assertIn('hello', self.x)
        self.assertIn('hello:thing', self.x)

    def test_not_in(self):
        self.assertNotIn('nothing', self.x)

    def test_not_in_when_unset(self):
        self.assertNotIn('error', self.x)
        class SubInjector(BasicInjector):
            pass
        @SubInjector.factory('picky')
        def no_spam(name=None):
            if name and 'spam' in name:
                raise jeni.UnsetError()
            elif name:
                return name
            else:
                return "I don't like spam!"
        x = jeni.InjectorProxy(SubInjector())
        self.assertIn('picky', x)
        self.assertIn('picky:foo', x)
        self.assertNotIn('picky:spamspamspam', x)

    def test_class(self):
        self.assertRaises(TypeError, jeni.InjectorProxy, BasicInjector)


class TestClassInProgress(unittest.TestCase):
    def test_class_in_progress(self):
        class Dummy(object):
            self.assertTrue(jeni.class_in_progress())

    def test_class_not_in_progress(self):
        self.assertFalse(jeni.class_in_progress())

    def test_broken_stack(self):
        stack = [(None, None, None, None, None, None)]
        self.assertFalse(jeni.class_in_progress(stack=stack))


if __name__ == '__main__': unittest.main()



# jeni.py
# Copyright 2013-2015 Ron DuPlain <ron.duplain@gmail.com> (see AUTHORS file).
# Released under the BSD License (see LICENSE file).

"""``jeni`` injects annotated dependencies"""

__version__ = '0.4.0'

import abc
import collections
import functools
import inspect
import re
import sys

import six


MAYBE = 'maybe'
PARTIAL = 'partial'
PARTIAL_REGARDLESS = 'partial_regardless'
EAGER_PARTIAL = 'eager_partial'
EAGER_PARTIAL_REGARDLESS = 'eager_partial_regardless'
WRAPPER_ASSIGNMENTS = functools.WRAPPER_ASSIGNMENTS + ('__notes__',)



class UnsetError(LookupError):
    """Note is not able to be provided, as it is currently unset."""
    def __init__(self, *a, **kw):
        self.note = kw.pop('note', None)
        super(UnsetError, self).__init__(*a, **kw)


class DependencyCycleError(RuntimeError):
    """Note is not able to be provided, because it depends on itself."""
    def __init__(self, *a, **kw):
        self.notes = kw.pop('notes', None)
        super(DependencyCycleError, self).__init__(*a, **kw)


@six.add_metaclass(abc.ABCMeta)
class Provider(object):
    """Provide a single prepared dependency."""

    @abc.abstractmethod
    def get(self, name=None):
        """Implement in subclass.

        Annotations in the form of ``'object:name'`` will pass the `name` value
        to the `get` method of the registered `Provider` (in this case, the
        provider registered with the `Injector` to provide `object`). This
        get-by-name pattern is useful for providers which have a dependency
        which supports lookups by key (e.g. HTTP headers or records in a
        key-value store).
        """

    def close(self):
        """By default, does nothing. Close objects as needed in subclass.

        Provider close methods should not intentionally raise errors.
        Specifically, if a dependency has transactions, the transaction should
        be committed or rolled back before close is called, and not left as an
        operation to be called during the close phase.

        Provider close methods must not take an argument; an injector cannot
        apply provided values on a close method since some providers may have
        already been closed. If an injected value is needed for the close
        method, annotate ``__init__`` and access the value via `self`.
        """


class FactoryProvider(Provider):
    """Adapt factory functions to the Provider interface.

    `Injector` uses this class to support registering factories.
    """
    unset_error = None

    @classmethod
    def bind(cls, fn):
        @annotate(annotate.partial_regardless(fn))
        def init(fn):
            return cls(fn)
        return init

    def __init__(self, function):
        self.function = function
        try:
            self.value = function()
        except UnsetError as err:
            self.unset_error = err

    def get(self, name=None):
        if name is not None:
            return self.function(name)
        if self.unset_error is not None:
            raise self.unset_error
        return self.value


class GeneratorProvider(Provider):
    """Manage generator lifecycle to implement Provider interface.

    `Injector` uses this class to support registering generators.
    """

    @classmethod
    def bind(cls, fn, support_name=False):
        @annotate(annotate.partial_regardless(fn))
        def init(fn):
            return cls(fn, support_name=support_name)
        return init

    def __init__(self, function, support_name=False):
        """Accept generator function & whether generator supports send."""
        self.function = function
        self.support_name = support_name

        self.generator = function()
        if not inspect.isgenerator(self.generator):
            msg = '{!r} is not a generator function'
            raise TypeError(msg.format(function))

        try:
            self.init_value = next(self.generator)
        except StopIteration:
            msg = "generator didn't yield: function {!r}"
            raise RuntimeError(msg.format(self.function))

    def get(self, name=None):
        """Get initial yield value, or result of send(name) if name given."""
        if name is None:
            return self.init_value
        elif not self.support_name:
            msg = "generator does not support get-by-name: function {!r}"
            raise TypeError(msg.format(self.function))
        try:
            value = self.generator.send(name)
        except StopIteration:
            msg = "generator didn't yield: function {!r}"
            raise RuntimeError(msg.format(self.function))
        return value

    def close(self):
        """Close the generator."""
        if self.support_name:
            self.generator.close()
        try:
            next(self.generator)
        except StopIteration:
            return
        else:
            msg = "generator didn't stop: function {!r}"
            raise RuntimeError(msg.format(self.function))


def see_doc(obj_with_doc):
    """Copy docstring from existing object to the decorated callable."""
    def decorator(fn):
        fn.__doc__ = obj_with_doc.__doc__
        return fn
    return decorator


class Annotator(object):
    """Class intent: serve as a stateless dict of function pointers.

    Annotate callables, settings data on callable objects themselves,
    providing hints for modes like maybe and partial.

    Annotations on callables are data for jeni's injection.
    Built as a class to embed annotation helpers and support customization.
    """

    def __call__(self, *notes, **keyword_notes):
        """Annotate a callable with a decorator to provide data for Injectors.

        Intended use::

            from jeni import annotate

            @annotate('foo', 'bar')
            def function(foo, bar):
                return

        An `Injector` would then need to register providers for 'foo' and 'bar'
        in order to apply this function; an injector with such providers can
        apply the annotated function without any further information::

            injector.apply(function)

        To get a partially applied function, to call later::

            fn = injector.partial(function)
            fn()

        Annotation does not alter the callable's default behavior.
        Call it normally::

            foo, bar = 'foo', 'bar'
            function(foo, bar)

        On Python 2, use decorators to annotate.
        On Python 3, use either decorators or function annotations::

            from jeni import annotate

            @annotate
            def function(foo: 'foo', bar: 'bar'):
                return

        Note that when using Python function annotations, all injected values
        are provided as keyword arguments.

        Since function annotations could be interpreted differently by
        different packages, injectors do not use ``function.__annotations__``
        directly. Functions opt in by a simple ``@annotate``
        decoration. Functions with Python annotations which have not been
        decorated are assumed to not be decorated for injection.

        (For this reason, annotating a callable with a single note where the
        note is a callable is not supported.)

        Notes which are provided to `annotate` (above 'foo' and 'bar') can be
        any hashable object (i.e. object able to be used as a key in a dict)
        and is not limited to strings. If tuples are used as notes, they must
        be of length 2, and `('maybe', ...)` and `('partial', ...)` are
        reserved.
        """
        if not keyword_notes and len(notes) == 1 and callable(notes[0]):
            # Here @annotate is being used without arguments.
            fn = notes[0]
            if not hasattr(fn, '__annotations__'):
                msg = '{!r} does not have annotations'
                raise AttributeError(msg.format(fn))
            self.set_annotations(fn, **fn.__annotations__)
            return fn
        def decorator(__fn):
            self.set_annotations(__fn, *notes, **keyword_notes)
            return __fn
        return decorator

    # When getting or setting annotations, check callable for __func__. If
    # found, the callable is a method, and the __func__ as function object
    # should be used instead.

    @classmethod
    def get_annotations(cls, __fn):
        """Get the annotations of a given callable."""
        if hasattr(__fn, '__func__'):
            __fn = __fn.__func__
        if hasattr(__fn, '__notes__'):
            return __fn.__notes__
        raise AttributeError('{!r} does not have annotations'.format(__fn))

    @classmethod
    def set_annotations(cls, __fn, *notes, **keyword_notes):
        """Set the annotations on the given callable."""
        if hasattr(__fn, '__func__'):
            __fn = __fn.__func__
        if hasattr(__fn, '__notes__'):
            msg = 'callable already has notes: {!r}'
            raise AttributeError(msg.format(__fn))
        __fn.__notes__ = (notes, keyword_notes)

    @classmethod
    def has_annotations(cls, __fn):
        """True if callable is annotated, else False."""
        try:
            cls.get_annotations(__fn)
        except AttributeError:
            return False
        return True

    @staticmethod
    def wraps(__fn, **kw):
        """Like ``functools.wraps``, with support for annotations."""
        kw['assigned'] = kw.get('assigned', WRAPPER_ASSIGNMENTS)
        return functools.wraps(__fn, **kw)

    @staticmethod
    def maybe(note):
        """Wrap a keyword note to record that its resolution is optional.

        Normally all annotations require fulfilled dependencies, but if a
        keyword argument is annotated as `maybe`, then on apply, an injector
        does not attempt to pass dependencies which are unset or not provided::

            from jeni import annotate

            @annotate('foo', bar=annotate.maybe('bar'))
            def foobar(foo, bar=None):
                return
        """
        return (MAYBE, note)

    @staticmethod
    def partial(__fn, *a, **kw):
        """Wrap a note for injection of a partially applied function.

        This allows for annotated functions to be injected for composition::

            from jeni import annotate

            @annotate('foo', bar=annotate.maybe('bar'))
            def foobar(foo, bar=None):
                return

            @annotate('foo', annotate.partial(foobar))
            def bazquux(foo, fn):
                # fn: injector.partial(foobar)
                return

        Keyword arguments are treated as `maybe` when using partial, in order
        to allow partial application of only the notes which can be provided,
        where the caller could then apply arguments known to be unavailable in
        the injector. Note that with Python 3 function annotations, all
        annotations are injected as keyword arguments.

        Injections on the partial function are lazy and not applied until the
        injected partial function is called. See `eager_partial` to inject
        eagerly.
        """
        return (PARTIAL, (__fn, a, tuple(kw.items())))

    @staticmethod
    def partial_regardless(__fn, *a, **kw):
        """Wrap a note for injection of a partially applied function, or don't.

        Use this instead of `partial` when binding a callable that may or may
        not have annotations.
        """
        return (PARTIAL_REGARDLESS, (__fn, a, tuple(kw.items())))

    @staticmethod
    def eager_partial(__fn, *a, **kw):
        """Wrap a note for injection of an eagerly partially applied function.

        Use this instead of `partial` when eager injection is needed in place
        of lazy injection.
        """
        return (EAGER_PARTIAL, (__fn, a, tuple(kw.items())))

    @staticmethod
    def eager_partial_regardless(__fn, *a, **kw):
        """Wrap a note for injection of an eagerly partially applied function, or don't.

        Use this instead of `eager_partial partial` when binding a callable
        that may or may not have annotations.
        """
        return (EAGER_PARTIAL_REGARDLESS, (__fn, a, tuple(kw.items())))

annotate = Annotator()
wraps = annotate.wraps
maybe = annotate.maybe
partial = annotate.partial
eager_partial = annotate.eager_partial


class Injector(object):
    """Collects dependencies and reads annotations to inject them."""
    annotator_class = Annotator
    factory_provider = FactoryProvider
    generator_provider = GeneratorProvider
    re_note = re.compile(r'^(.*?)(?::(.*))?$') # annotation is 'object:name'

    def __init__(self, provide_self=False):
        """A subclass could take arguments, but should pass keywords to super.

        An Injector subclass inherits the provider registry of its base
        classes, but can override any provider by re-registering notes. When
        organizing a project, create an Injector subclass to serve as the
        object to register all providers. This allows for the project to have
        its own namespace of registered dependencies. This registry can be
        customized by further subclasses, either for injecting mocks in testing
        or providing alternative dependencies in a different runtime::

            from jeni import Injector as BaseInjector

            class Injector(BaseInjector):
                "Subclass provides namespace when registering providers."

        By default, the injector does not provide itself, but will when asked::

            injector = Injector(provide_self=True)
            injector.get('injector')

        This is useful in a context manager::

            with Injector(provide_self=True) as injector:
                injector.get('injector')

        Annotate with note 'injector' to inject the injector.
        """
        if provide_self:
            self.value('injector', self)

        self.annotator = self.annotator_class()

        self.closed = False
        self.instances = {}
        self.values = {}

        self.finalizers = []

        #: Statistics for resolved notes, note -> count.
        #: Records counts as soon as get is called, even if unset or error.
        self.stats = collections.defaultdict(int)

        #: Collection of note tuples which are currently being instantiated.
        #: This allows for dependency cycle checks.
        self.instantiating = []

    @classmethod
    def provider(cls, note, provider=None, name=False):
        """Register a provider, either a Provider class or a generator.

        Provider class::

            from jeni import Injector as BaseInjector
            from jeni import Provider

            class Injector(BaseInjector):
                pass

            @Injector.provider('hello')
            class HelloProvider(Provider):
                def get(self, name=None):
                    if name is None:
                        name = 'world'
                    return 'Hello, {}!'.format(name)

        Simple generator::

            @Injector.provider('answer')
            def answer():
                yield 42

        If a generator supports get with a name argument::

            @Injector.provider('spam', name=True)
            def spam():
                count_str = yield 'spam'
                while True:
                    count_str = yield 'spam' * int(count_str)

        Registration can be a decorator or a direct method call::

            Injector.provider('hello', HelloProvider)
        """
        def decorator(provider):
            if inspect.isgeneratorfunction(provider):
                # Automatically adapt generator functions
                provider = cls.generator_provider.bind(
                        provider, support_name=name)
                return decorator(provider)

            cls.register(note, provider)
            return provider

        if provider is not None:
            decorator(provider)
        else:
            return decorator

    @classmethod
    def factory(cls, note, fn=None):
        """Register a function as a provider.

        Function (name support is optional)::

            from jeni import Injector as BaseInjector
            from jeni import Provider

            class Injector(BaseInjector):
                pass

            @Injector.factory('echo')
            def echo(name=None):
                return name

        Registration can be a decorator or a direct method call::

            Injector.factory('echo', echo)
        """
        def decorator(f):
            provider = cls.factory_provider.bind(f)
            cls.register(note, provider)
            return f

        if fn is not None:
            decorator(fn)
        else:
            return decorator

    @classmethod
    def value(cls, note, scalar):
        """Register a single value to be provided.

        Supports base notes only, does not support get-by-name notes.
        """
        cls.factory(note, lambda: scalar)

    def apply(self, fn, *a, **kw):
        """Fully apply annotated callable, returning callable's result."""
        args, kwargs = self.prepare_callable(fn)
        args += a; kwargs.update(kw)
        return fn(*args, **kwargs)

    def partial(self, fn, *user_args, **user_kwargs):
        """Return function with closure to lazily inject annotated callable.

        Repeat calls to the resulting function will reuse injections from the
        first call.

        Positional arguments are provided in this order:

        1. positional arguments provided by injector
        2. positional arguments provided in `partial_fn = partial(fn, *args)`
        3. positional arguments provided in `partial_fn(*args)`

        Keyword arguments are resolved in this order (later override earlier):

        1. keyword arguments provided by injector
        2. keyword arguments provided in `partial_fn = partial(fn, **kwargs)`
        3. keyword arguments provided in `partial_fn(**kargs)`

        Note that Python function annotations (in Python 3) are injected as
        keyword arguments, as documented in `annotate`, which affects the
        argument order here.

        `annotate.partial` accepts arguments in same manner as this `partial`.
        """
        self.get_annotations(fn) # Assert has annotations.
        def lazy_injection_fn(*run_args, **run_kwargs):
            arg_pack = getattr(lazy_injection_fn, 'arg_pack', None)
            if arg_pack is not None:
                pack_args, pack_kwargs = arg_pack
            else:
                jeni_args, jeni_kwargs = self.prepare_callable(fn, partial=True)
                pack_args = jeni_args + user_args
                pack_kwargs = {}
                pack_kwargs.update(jeni_kwargs)
                pack_kwargs.update(user_kwargs)
                lazy_injection_fn.arg_pack = (pack_args, pack_kwargs)
            final_args = pack_args + run_args
            final_kwargs = {}
            final_kwargs.update(pack_kwargs)
            final_kwargs.update(run_kwargs)
            return fn(*final_args, **final_kwargs)
        return lazy_injection_fn

    def eager_partial(self, fn, *a, **kw):
        """Partially apply annotated callable, returning a partial function.

        By default, `partial` is lazy so that injections only happen when they
        are needed. Use `eager_partial` in place of `partial` when a guarantee
        of injection is needed at the time the partially applied function is
        created.

        `eager_partial` resolves arguments similarly to `partial` but relies on
        `functools.partial` for argument resolution when calling the final
        partial function.
        """
        args, kwargs = self.prepare_callable(fn, partial=True)
        args += a; kwargs.update(kw)
        return functools.partial(fn, *args, **kwargs)

    def apply_regardless(self, fn, *a, **kw):
        """Like `apply`, but applies if callable is not annotated."""
        if self.has_annotations(fn):
            return self.apply(fn, *a, **kw)
        return fn(*a, **kw)

    def partial_regardless(self, fn, *a, **kw):
        """Like `partial`, but applies if callable is not annotated."""
        if self.has_annotations(fn):
            return self.partial(fn, *a, **kw)
        else:
            return functools.partial(fn, *a, **kw)

    def eager_partial_regardless(self, fn, *a, **kw):
        """Like `eager_partial`, but applies if callable is not annotated."""
        if self.has_annotations(fn):
            return self.eager_partial(fn, *a, **kw)
        return functools.partial(fn, *a, **kw)

    def get(self, note):
        """Resolve a single note into an object."""
        if self.closed:
            raise RuntimeError('{!r} already closed'.format(self))

        # Record request for note even if it fails to resolve.
        self.stats[note] += 1

        # Handle injection of partially applied annotated functions.
        if isinstance(note, tuple) and len(note) == 2:
            if note[0] == PARTIAL:
                fn, a, kw_items = note[1]
                return self.partial(fn, *a, **dict(kw_items))
            elif note[0] == PARTIAL_REGARDLESS:
                fn, a, kw_items = note[1]
                return self.partial_regardless(fn, *a, **dict(kw_items))
            elif note[0] == EAGER_PARTIAL:
                fn, a, kw_items = note[1]
                return self.eager_partial(fn, *a, **dict(kw_items))
            elif note[0] == EAGER_PARTIAL_REGARDLESS:
                fn, a, kw_items = note[1]
                return self.eager_partial_regardless(fn, *a, **dict(kw_items))

        basenote, name = self.parse_note(note)
        if name is None and basenote in self.values:
            return self.values[basenote]
        try:
            provider_factory = self.lookup(basenote)
        except LookupError:
            msg = "Unable to resolve '{}'"
            raise LookupError(msg.format(note))

        self.instantiating.append((basenote, name))
        try:
            if self.instantiating.count((basenote, name)) > 1:
                stack = ' <- '.join(repr(note) for note in self.instantiating)
                notes = tuple(self.instantiating)
                raise DependencyCycleError(stack, notes=notes)

            return self.handle_provider(provider_factory, note)
        finally:
            self.instantiating.pop()

    def close(self):
        """Close injector & injected Provider instances, including generators.

        Providers are closed in the reverse order in which they were opened,
        and each provider is only closed once. Providers are closed if accessed
        by the injector, even if a dependency is not successfully provided. As
        such, providers should determine whether or not anything needs to be
        done in the close method.
        """
        if self.closed:
            raise RuntimeError('{!r} already closed'.format(self))
        for finalizer in reversed(self.finalizers):
            # Note: Unable to apply injector on close method.
            finalizer()
        self.closed = True

    def prepare_callable(self, fn, partial=False):
        """Prepare arguments required to apply function."""
        notes, keyword_notes = self.get_annotations(fn)
        return self.prepare_notes(*notes, __partial=partial, **keyword_notes)

    def prepare_notes(self, *notes, **keyword_notes):
        """Get injection values for all given notes."""
        __partial = keyword_notes.pop('__partial', False)
        args = tuple(self.get(note) for note in notes)
        kwargs = {}
        for arg in keyword_notes:
            note = keyword_notes[arg]
            if isinstance(note, tuple) and len(note) == 2 and note[0] == MAYBE:
                try:
                    kwargs[arg] = self.get(note[1])
                except LookupError:
                    continue
            elif __partial:
                try:
                    kwargs[arg] = self.get(note)
                except LookupError:
                    continue
            else:
                kwargs[arg] = self.get(note)
        return args, kwargs

    @classmethod
    def parse_note(cls, note):
        """Parse string annotation into object reference with optional name."""
        if isinstance(note, tuple):
            if len(note) != 2:
                raise ValueError('tuple annotations must be length 2')
            return note
        try:
            match = cls.re_note.match(note)
        except TypeError:
            # Note is not a string. Support any Python object as a note.
            return note, None
        return match.groups()

    def handle_provider(self, provider_factory, note):
        """Get value from provider as requested by note."""
        # Implementation in separate method to support accurate book-keeping.
        basenote, name = self.parse_note(note)

        # _handle_provider could be even shorter if
        # Injector.apply() worked with classes, issue #9.
        if basenote not in self.instances:
            if (isinstance(provider_factory, type) and
                    self.has_annotations(provider_factory.__init__)):
                args, kwargs = self.prepare_callable(provider_factory.__init__)
                self.instances[basenote] = provider_factory(*args, **kwargs)

            else:
                self.instances[basenote] = self.apply_regardless(
                        provider_factory)

            provider = self.instances[basenote]
            if hasattr(provider, 'close'):
                self.finalizers.append(self.instances[basenote].close)

        provider = self.instances[basenote]
        get = self.partial_regardless(provider.get)

        try:
            if name is not None:
                return get(name=name)
            self.values[basenote] = get()
            return self.values[basenote]

        except UnsetError:
            # Use sys.exc_info to support both Python 2 and Python 3.
            exc_type, exc_value, tb = sys.exc_info()
            exc_msg = str(exc_value)
            if exc_msg:
                msg = '{}: {!r}'.format(exc_msg, note)
            else:
                msg = repr(note)
            six.reraise(exc_type, exc_type(msg, note=note), tb)

    @classmethod
    def register(cls, note, provider):
        """Implementation to register provider via `provider` & `factory`."""
        basenote, name = cls.parse_note(note)
        if 'provider_registry' not in vars(cls):
            cls.provider_registry = {}
        cls.provider_registry[basenote] = provider

    @classmethod
    def lookup(cls, basenote):
        """Look up note in registered annotations, walking class tree."""
        # Walk method resolution order, which includes current class.
        for c in cls.mro():
            if 'provider_registry' not in vars(c):
                # class is a mixin, super to base class, or never registered.
                continue
            if basenote in c.provider_registry:
                # note is in the registry.
                return c.provider_registry[basenote]
        raise LookupError(repr(basenote))

    def __enter__(self):
        """Support for context manager, returning self."""
        return self

    def enter(self):
        """Enter context-manager without with-block. See also: `exit`.

        Useful for before- and after-hooks which cannot use a with-block.
        """
        return self.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        """Support for context manager, close on exit."""
        self.close()

    def exit(self):
        """Exit context-manager without with-block. See also: `enter`."""
        return self.__exit__(None, None, None)

    @see_doc(Annotator.get_annotations)
    def get_annotations(self, *a, **kw):
        return self.annotator.get_annotations(*a, **kw)

    @see_doc(Annotator.has_annotations)
    def has_annotations(self, *a, **kw):
        return self.annotator.has_annotations(*a, **kw)

    @classmethod
    def sub(cls, *mixins_and_dicts, **values):
        """Create and instantiate a sub-injector.

        Mixins and local value dicts can be passed in as arguments.  Local
        values can also be passed in as keyword arguments.
        """

        class SubInjector(cls):
            pass

        mixins = [ x for x in mixins_and_dicts if isinstance(x, type) ]
        if mixins:
            SubInjector.__bases__ = tuple(mixins) + SubInjector.__bases__

        dicts = [ x for x in mixins_and_dicts if not isinstance(x, type) ]
        for d in reversed(dicts):
            for k,v in d.items():
                if k not in values:
                    values[k] = v

        for k,v in values.items():
            SubInjector.value(k, v)

        return SubInjector()


class InjectorProxy(object):
    """Forwards getattr & getitem to enclosed injector.

    If an injector has 'hello' registered::

        from jeni import InjectorProxy
        deps = InjectorProxy(injector)
        deps.hello

    Get by name can use dict-style access::

        deps['hello:name']
    """

    def __init__(self, injector):
        if inspect.isclass(injector):
            msg = 'takes an instance not a class, {!r}'
            raise TypeError(msg.format(injector))
        self.injector = injector

    def __getattr__(self, name):
        return self.injector.get(name)

    def __getitem__(self, key):
        return self.injector.get(key)

    def __contains__(self, item):
        try:
            self.injector.get(item)
        except LookupError:
            return False
        return True


def class_in_progress(stack=None):
    """True if currently inside a class definition, else False."""
    if stack is None:
        stack = inspect.stack()
    for frame in stack:
        statement_list = frame[4]
        if statement_list is None:
            continue
        if statement_list[0].strip().startswith('class '):
            return True
    return False
